#include "bayescc2.h"
#include "common.h"
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/uniform.hpp>


using namespace boost;

// interesting and confusing: http://code.google.com/p/ourmine/wiki/LectureNaiveBayes


BayesCC2::BayesCC2(Graph * _graph, double _alpha, double _beta, double _teta) :
        graph(_graph), alpha(_alpha), beta(_beta), teta(_teta)
{    
    
    cout << "#INFO alpha= " << alpha << " beta= " << beta << " teta= " << teta <<"\n";


    K = 1; // alpha as in http://en.wikipedia.org/wiki/Additive_smoothing
    M = 2; // m as in m-estimate: http://www.coli.uni-saarland.de/~crocker/courses/learning/lecture10.pdf


    nLabels = graph->nLabels;

    changed = true;

    assert(alpha >= 0 && alpha <= 1);
    assert(beta >= 0 && beta <= 1);
    assert(teta >= 0 && teta <= 1);

}



BayesCC2::~BayesCC2() {
}





void BayesCC2::labelRemainingNodesUsingPriors() {

    int count = 0;

    int topLabel = topPrior();

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(KNOWN_LABEL(node.labelId))
            continue;

        //   cout << "WARNING: labelRemainingNodesUsingPriors on node id=" << node.id << "\n";

        node.labelId = topLabel;
        node.labelTime = iterationCount;

        count++;

    }

    cout << "Labeled " << count << " nodes using priors.\n";

}

set<VertexID> BayesCC2::lookupInitialCandidates()  {
    set<VertexID> candidates;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];


        if(KNOWN_LABEL(node.labelId)) {

            MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
            boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices(vertexID, graph->g);

            for (; neighbourIt != neighbourEnd; ++neighbourIt){
                VertexID neighborID = *neighbourIt;

                
                Vertex & neighbor = graph->g[neighborID];

                if(UNKNOWN_LABEL(neighbor.labelId)) {
                    candidates.insert(*neighbourIt);
                    //                 cout << "adding initial candidate node id=" << neighbor.id <<"\n";
                }

            }
        }
    }

    // the count of initial candidates does not change for any vale of beta < 1.    
    cout << "Count of initial candidate nodes: "<< candidates.size() << "\n";

    return candidates;
}





double BayesCC2::naiveBayesP(int labelId, VertexID vertexID) {
    double P = labelPriorsP[labelId];

    MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
    boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( vertexID, graph->g );

    for (; neighbourIt != neighbourEnd; ++neighbourIt){

        VertexID neighborID = *neighbourIt;
        Vertex & neighbor = graph->g[neighborID];

        if(UNKNOWN_LABEL(neighbor.labelId))
            continue;

        assert(KNOWN_LABEL(neighbor.labelId));
        assert(labelJointP[neighbor.labelId][labelId] != 0);

        P *= labelJointP[neighbor.labelId][labelId];        
    }

    return P;
}



void BayesCC2::labelNode(VertexID vertexID) {

    Vertex & node = graph->g[vertexID];

    assert(node.labelId == 0); // we never label nodes again.

    assert(node.training == false); // we never ever label training nodes.

    int nLabels = graph->labels.size();
    double histogram[nLabels+1];
    double sum = 0;
    for(int i = 1; i <= nLabels; i++) {
        histogram[i] = naiveBayesP(i, vertexID);
        sum += histogram[i];
    }

    if(sum > 0)
        for(int i = 1; i <= nLabels; i++) {
        histogram[i] /= sum; // normalization..
        pcheck(histogram[i]);
    }

    int topLabel = 0;
    double topP = 0;

    for(int i = 1; i <= nLabels; i++) {
        if(histogram[i] > topP) {
            topP = histogram[i];
            topLabel = i;
        }
    }

    if(topLabel == 0) {
        topLabel = topPrior();
        topP = labelPriorsP[topLabel];        
    }

    assert(  !isnan(topP) && !isinf(topP));

    assert(topLabel != 0);

    changed |= node.labelId == topLabel;

    if(node.labelId != topLabel)
        node.labelTime = iterationCount;

    if(node.labelId != 0 && node.labelId != topLabel)
        nChanges++;

    if(node.labelId == 0)
        nLabeled++;

    assert(  !isnan(topP) && !isinf(topP));


    node.labelId = topLabel;

}


int BayesCC2::topPrior() {

    int topLabelId = 0;
    double topP = 0;

    for(unsigned int labelId = 1; labelId < labelPriorsP.size(); labelId++) {
        if(labelPriorsP[labelId] > topP) {
            topLabelId = labelId;
            topP = labelPriorsP[labelId];
        }
    }

    return topLabelId;

}

void BayesCC2::updateJointProbabilities() {

    labelJointP.clear();
    labelJointP.resize(graph->labels.size()+1);
    for(unsigned int i = 0; i <= graph->labels.size(); i++) {
        labelJointP[i].resize(graph->labels.size()+1, 0);
    }

    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(graph->g);

    double counts[graph->labels.size()+1];

    for(unsigned int i = 1; i <= graph->labels.size(); i++)
        counts[i] = 0;

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = graph->g[edgeID];

        Vertex &node1 = graph->nodeFID(edge.node1);
        Vertex &node2 = graph->nodeFID(edge.node2);

        if(UNKNOWN_LABEL(node1.labelId) || UNKNOWN_LABEL(node2.labelId))
            continue;

        counts[node1.labelId]+= edge.P;
        counts[node2.labelId]+= edge.P;
        labelJointP[node1.labelId][node2.labelId]+= edge.P;
        labelJointP[node2.labelId][node1.labelId]+= edge.P;
    }


    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        for(unsigned int j = 1; j <= graph->labels.size(); j++) {

            double count = labelJointP[i][j];
            double prior = pcheck(labelPriorsP[i]);

            if(count > 0)  {
                labelJointP[i][j] = pcheck((count+(M*prior)) / (counts[j]+M));
            } else {
                labelJointP[i][j] = prior;
            }

            pcheck(labelJointP[i][j]);

        }
    }

}

inline double BayesCC2::pcheck(double P) {


    assert(!isnan(P));
    assert(!isinf(P));


    if(P >1)  {
        assert(P <= 1+1e-5);
        P = 1;
    }

    if(P < 0) {
        assert(P >= -1e-5);
        P = 0;
    }

    return P;


}




bool BayesCC2::iterate(set<VertexID> &candidates) {

    for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {
        labelNode(*iter);
    }

    // build new list of reachable-nodes
    set<VertexID> newCandidates;
    lookupNextCandidates(candidates, newCandidates);

    candidates = newCandidates;

    return (candidates.size() >0);
}

void BayesCC2::lookupNextCandidates(set<VertexID> &candidates, set<VertexID> &newCandidates) {

    for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {

        MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
        boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( *iter, graph->g );

        for (; neighbourIt != neighbourEnd; ++neighbourIt){
            VertexID neighborID = *neighbourIt;

            Vertex & neighbor = graph->g[neighborID];
            if(UNKNOWN_LABEL(neighbor.labelId)) {
                newCandidates.insert(neighborID);
            }
        }

    }
}

void BayesCC2::updateLabelPriorsP() {

    labelPriorsP.resize(graph->labels.size()+1);

    for(unsigned int labelId = 1; labelId < labelPriorsP.size(); labelId++) {
        labelPriorsP[labelId] = 0;
    }

    double count = 0;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(UNKNOWN_LABEL(node.labelId))
            continue;

        count++;
        labelPriorsP[node.labelId]+= 1.0;
    }

    for(unsigned int labelId = 1; labelId < labelPriorsP.size(); labelId++) {
        labelPriorsP[labelId] = (labelPriorsP[labelId]+K) / (count+K*nLabels);
        pcheck(labelPriorsP[labelId]);
    }


}






void BayesCC2::eval() {
    
    elapsedTime.begin();

    nLabeled = 0;
    iterationCount = 1;

    //   cout << "timee1: " << elapsedTime.get() << "\n";


    set<VertexID> candidates = lookupInitialCandidates();

    //   cout << "timee3: " << elapsedTime.get() << "\n";

    // update prior probabilities
    updateLabelPriorsP();

    // update joint probabilies
    updateJointProbabilities();

    //   cout << "timee4: " << elapsedTime.get() << "\n";

    for(;;) {
        nChanges = 0;

        int c = candidates.size();

        bool ret = iterate(candidates);

        double ratioLabeled = ((double)nLabeled) / ((double)num_vertices(graph->g));
        double ratioCandidates = ((double)c) / ((double)num_vertices(graph->g));

        cout << "#ITER n= " << iterationCount << " c= " << c << " ratioCandidates= " << ratioCandidates << " cNext= " << candidates.size() << " nLabeled= " << nLabeled << " nChanges= " << nChanges  << " ratioLabeled= " << ratioLabeled <<  " Acc= " << Evaluation::testAccuracy(graph) <<  " iAcc= " << Evaluation::testAccuracy(graph, iterationCount) <<" time= " << elapsedTime.end(false) << "\n";

        iterationCount++;

        if(ret == false)
            break;
    }

    labelRemainingNodesUsingPriors();

}

