#include "samplingcc.h"
#include "graph.h"
#include "common.h"
#include "elapsedtime.h"
#include <set>
#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/uniform.hpp>
#include <boost/random.hpp>

#include "evaluation.h"

#include "random.h"

using namespace boost;
using namespace std;


// we use eLabelId as input
// we put the assignment determined by algorithm in labelId


SamplingCC::SamplingCC(Graph * _graph, int _nSamples) :
        graph(_graph), nSamples(_nSamples)
{

    labelIds = NULL;
    init();
}


SamplingCC::~SamplingCC() {

    SAFE_FREE(labelIds);

    /*
    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];
        node.labelIds.clear();
    }*/

}


void SamplingCC::init() {

    assert(labelIds == NULL);

    long labelIdsSize = (graph->labels.size()+1) * num_vertices(graph->g)* sizeof( long);
    labelIds = ( long *) malloc(labelIdsSize);
    memset(labelIds, 0, labelIdsSize);

    /*
        MyGraph::vertex_iterator vertexIt, vertexEnd;
        boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
        for (; vertexIt != vertexEnd; ++vertexIt){
            VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
            Vertex & node = graph->g[vertexID];
            //node.labelIds.resize(nSamples, 0);
        }
    */

}


bool SamplingCC::iterate(set<VertexID> &candidates) {
    //   cout <<  "New iteration, labeling " << candidates.size() << " nodes ...\n";

    for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {
        labelNode(*iter);
    }

    set<VertexID> newCandidates;
    lookupNextCandidates(candidates, newCandidates);

    candidates = newCandidates;

    // do we need more iterations?
    return (candidates.size() > 0);

}



void SamplingCC::labelNode(VertexID vertexID) {

    Vertex & node = graph->g[vertexID];

    //  cout << "Labeling node id=" << node.id << "\n";


    int nLabels = graph->labels.size()+1; // +1 for UNKNOWN_LABEL, 0.
    double histogram[nLabels]; // xxx here we assume that label ids are sequential, starting from 1.
    for(unsigned int i = 1; i <= graph->labels.size(); i++)
        histogram[i] = 0;

    MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
    boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( vertexID, graph->g );

    for (; neighbourIt != neighbourEnd; ++neighbourIt){
        VertexID neighborID = *neighbourIt;

        if(!graph->edge(vertexID,neighborID).active)
            continue;

        Vertex & neighbor = graph->g[neighborID];

        if(UNKNOWN_LABEL(neighbor.labelId))
            continue;

        assert(neighbor.labelId < nLabels);

        histogram[neighbor.labelId]++;
    }


    int topLabel = 0;
    double topFreq = 0;

    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        if(histogram[i] > topFreq) {
            topFreq = histogram[i];
            topLabel = i;
        }
    }

    assert(topLabel != 0);

    node.labelId = topLabel;

}



void SamplingCC::lookupNextCandidates(set<VertexID> &candidates, set<VertexID> &newCandidates) {

    for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {

        VertexID vertexID = *iter; // dereference vertexIt, get the ID

        MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
        boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( *iter, graph->g );

        for (; neighbourIt != neighbourEnd; ++neighbourIt){
            VertexID neighborID = *neighbourIt;

            if(!graph->edge(vertexID,neighborID).active)
                continue;

            Vertex & neighbor = graph->g[neighborID];
            if(UNKNOWN_LABEL(neighbor.labelId)) {
                //    cout << "Inserting node id=" << neighbor.id << " in new candidates\n";
                newCandidates.insert(neighborID);
            }
        }

    }
}





void SamplingCC::initSample(int sampleId) {

    // init labelId for this sample. we need this because
    // after the first iteration, we do not know anymore
    // the original label as assigned by the Evaluation.
    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];
        node.labelId = node.eLabelId;
    }

    // set active edges.

    rand48 rng =  Random::getInstance();

    int count = 0;

    uniform_real<> ur(0.0,1.0);


    variate_generator<rand48&,uniform_real<> > drawSample(rng, ur);


    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(graph->g);

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = graph->g[edgeID];
        if(drawSample() <= edge.P) {
            edge.active = true;
            count += edge.active;
        } else {
            edge.active = false;
        }
    }        

    //  cout << "Activated " << count << "/" << num_edges(graph->g) << " edges\n";

}



set<VertexID> SamplingCC::lookupInitialCandidates()  {
    set<VertexID> candidates;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(!KNOWN_LABEL(node.labelId))
            continue;

        MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
        boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices(vertexID, graph->g);

        for (; neighbourIt != neighbourEnd; ++neighbourIt){
            VertexID neighborID = *neighbourIt;

            if(!graph->edge(vertexID,neighborID).active)
                continue;

            Vertex & neighbor = graph->g[neighborID];

            if(UNKNOWN_LABEL(neighbor.labelId))
                candidates.insert(*neighbourIt);
        }

    }

    assert(candidates.size() > 0);

    //  cout << "Initial candidates:  "<< candidates.size() << "\n";

    return candidates;
}





void SamplingCC::saveLabels(int sampleId) {

    int nodePos = 0;
    int nLabels = graph->labels.size()+1;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        labelIds[nodePos*nLabels+node.labelId]++; // labelIds[nodeId][labelId] = frequency (unsigned int)

        //node.labelIds[sampleId] = node.labelId;

        nodePos++;
    }
}






void SamplingCC::eval() {
    return eval(0);
}

void SamplingCC::eval(unsigned long  tMax) {

    ElapsedTime eTime;
    eTime.begin();

    for(int i = 0; i < nSamples; i++) {

        sampleCount = i;
        iterationCount = 1;

        if( (i+1) % 50 == 0)
            cout << "Sample " << (sampleCount+1) << "/" << nSamples << " (" << sampleCount*100/nSamples << "\%)\n";

        initSample(i); // generate new sample, i.e., updates $active state on edges.
        set<VertexID> candidates = lookupInitialCandidates();

        for(;;) {
            double ret = iterate(candidates);
            iterationCount++;

            if(!ret)
                break;
        }

//        if( (i+1) % 50 == 0)
  //      cout << "Iteration time: " << eTime.end(false) << "\n";

        saveLabels(i);

        if(tMax >0 && eTime.end(false) >= tMax) {
            break;
        }

    }

    saveTopLabels();

}



void SamplingCC::saveTopLabels() {
    int nodePos = 0;
    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];
        node.labelId = assignTopLabel(nodePos);
        nodePos++;
    }

}

int SamplingCC::assignTopLabel(int nodePos) {

    int topLabel = 0;
    int topFreq = 0;

    int nLabels = graph->labels.size()+1;

    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        if(labelIds[nodePos*nLabels+i] > topFreq) {
            topFreq = labelIds[nodePos*nLabels+i];
            topLabel = i;
        }
    }

    return topLabel;

}

